﻿using FlyAdX.UI;
using FlyAdXBase;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.IO;
using System.IO.Ports;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace FlyAdXBase
{
    /// <summary>
    /// 数据包都以7E 开头
    /// 数据中，7E 会转义为 7D 5E;7D 会转义为 7D 5D
    /// </summary>
    public class FlyAd2021ServiceClient : IFlyAd2021
    {
        public event PropertyChangedEventHandler PropertyChanged;



        #region 参数
        public string Addr { get; set; } = "COM14";
        #endregion

        #region IFlyAd2021
        public DateTime Now { get; private set; }
        
        public UInt16 AD { get; private set; }
        public UInt16 AD2 { get; private set; }
        public UInt16 ENC1 { get; private set; }
        public UInt16 ENC2 { get; private set; }
        public UInt16 In { get; private set; }
        public UInt16 Out { get; private set; }


        /// <summary>
        /// 通过systick 判断复位次数，
        /// Now的复位次数
        /// </summary>
        public int ResetCnt { get; private set; }

        /// <summary>
        /// timegrid 事件; 版本2021: 0.2s 一次
        /// </summary>
        public event TimeGridEventHandler TimeGridEvent;



        /// <summary>
        ///  通讯速度 byte/s
        /// </summary>
        public double CommSpeed { get; private set; }

        /// <summary>
        /// 传输速度 单位 pack/s
        /// </summary>
        public double PackSpeed { get; private set; }

        /// <summary>
        /// 包出错次数
        /// </summary>
        public int ErrCnt { get; private set; }

        public event PackEventHandler RecPackEvent;
        public event PackEventHandler FuncSendPackEvent;
        public event PackEventHandler FuncRecPackEvent;
        public event PackEventHandler DataRecPackEvent;
        #endregion


        public bool IsRunning { get; private set; }


        byte last_systick = 0;
        int packCnt = 0;
        int recCnt = 0;
        /// <summary>
        /// 数据
        /// </summary>
        List<UInt16> ADs = new List<ushort>();
        List<UInt16> AD2s = new List<ushort>();
        List<UInt16> ENC1s = new List<ushort>();
        List<UInt16> ENC2s = new List<ushort>();
        List <UInt16> Ins = new List<ushort>();


        List<byte> currPack = new List<byte>();
        Stopwatch stopwatch = new Stopwatch();
        CancellationTokenSource cancellationTokenSource;

        IFlyAd2021Comm comm;
        public void Init()
        {
            Load();

            this.PropertyChanged += FlyAd2021ServiceClient_PropertyChanged;
        }

        private void FlyAd2021ServiceClient_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(Addr))
            {
                Save();    
            }
        }
        string filePath = "flyad2021.json";
        bool Save() {
            return FlyAd2021JsonDb.Save(this, filePath);
        }
        bool Load()
        {
            return FlyAd2021JsonDb.Load(this, filePath);
        }

        void createComm()
        {
            if (Addr.ToLower().StartsWith("com"))
                comm = new FlyAd2021Comm_SerialPort();
            else
                comm = new FlyAd2021Comm_TcpClient();

            comm.RecMsg += Comm_RecMsg;

            Misc.BindingOperations.SetBinding(comm, nameof(comm.IsRunning), this, nameof(IsRunning));
        }

        private void Comm_RecMsg(byte[] recBuf)
        {
            recCnt += recBuf.Count();
            Parse(recBuf);
        }

        public void Start()
        {
            if (comm != null && comm.IsRunning)
                return;

            if (comm == null)
            {
                createComm();
            }
            else
            {
                if ((Addr.ToLower().StartsWith("com") && (comm is FlyAd2021Comm_SerialPort)) ||
                    (!Addr.ToLower().StartsWith("com") && (comm is FlyAd2021Comm_TcpClient)))
                {
                    //不用重建
                }
                else
                {
                    createComm();
                }
            }

            comm.Addr = Addr;
            comm.Start();

            //启动线程，测量速度
            cancellationTokenSource = new CancellationTokenSource();
            packCnt = 0;
            recCnt = 0;
            Task.Factory.StartNew(MeasureTask, cancellationTokenSource.Token);
        }

        public void Stop() {
            comm.Stop();

            //停止线程
            cancellationTokenSource.Cancel();
        }
        async void MeasureTask() {

            stopwatch.Start();
            while (!cancellationTokenSource.IsCancellationRequested) {
                if (stopwatch.Elapsed > TimeSpan.FromSeconds(1)) {
                    //1秒均值
                    CommSpeed = recCnt / stopwatch.Elapsed.TotalSeconds;
                    PackSpeed = packCnt / stopwatch.Elapsed.TotalSeconds;
                    recCnt = 0;
                    packCnt = 0;
                    stopwatch.Restart();
                }
                await Task.Delay(100);
            }
            stopwatch.Stop();
        }

        void Parse(byte[] recBuf) {
            for (int i = 0; i < recBuf.Count(); i++)
            {
                if (recBuf[i] == 0x7e)
                {
                    //找到头了
                    //结束之前的包
                    if (currPack.Count > 0)
                    {
                        var pack = currPack.ToArray();
                        RecPackEvent?.Invoke(this, new PackEventArgs() { Time = DateTime.Now, Pack = pack });
                        ParsePack(pack);
                        currPack.Clear();
                        packCnt++;
                    }
                }
                currPack.Add(recBuf[i]);
            }
        }

        /// <summary>
        /// 包解析
        /// </summary>
        void ParsePack(byte[] pack)
        {
            //第0个肯定是7E
            if (pack.Count() == 1)
            {
                //只有一个 7E
                return;
            }

            //转义数据
            if (!ProtocolCommon.Pdu2Data(pack, out List<byte> datas)) {
                //异常
                return;
            }


            //解析后的 datas 没有了 帧包装 7E
            if (datas.Count() == 1) 
            {
                //只有一个字节，异常！！！
                ErrCnt++;
                return;
            }
            byte crc8 = Misc.CRC.CRC8(datas, 0, datas.Count() - 1);
            if (datas.Last() != crc8) {
                //CRC8 出错
                ErrCnt++;
                return;
            }

            byte B0 = datas[0];
            bool PT = Misc.MyBase.CHECKBIT(B0, 7);
            byte systick = B0;
            Misc.MyBase.CLEARBIT(ref systick, 7);

            if (Now == DateTime.MinValue)
            {
                Now = DateTime.Now;
                ResetCnt++;
            }
            else {
                int inc = CalSystickInc(systick);
                if (inc > 50)
                {
                    //非常异常，突然跳过了50ms
                    //复位
                    Clear();
                    Now = DateTime.Now;
                    ResetCnt++;
                }
                else {
                    Now += TimeSpan.FromMilliseconds(inc);
                }
            }
            
            last_systick = systick;
            if (!PT)
            {
                ParseDataPack(datas);
            }
            else {
                ParseFuncPack(datas);
            }
            
        }
        int CalSystickInc(byte systick) {
            int inc = systick - last_systick;
            if (inc < 0)
                inc += 128;
            return inc;
        }
        void ParseDataPack(List<byte> datas) 
        {
            DataRecPackEvent?.Invoke(this, new PackEventArgs() { Time = DateTime.Now, Pack = datas.ToArray() });

            var buf = datas.ToArray();

            byte B1 = datas[1];
            bool hasIn = Misc.MyBase.CHECKBIT(B1, 7);
            bool hasENC1 = Misc.MyBase.CHECKBIT(B1, 6);
            bool hasENC2 = Misc.MyBase.CHECKBIT(B1, 5);
            bool hasOut = Misc.MyBase.CHECKBIT(B1, 4);
            bool hasAd2 = Misc.MyBase.CHECKBIT(B1, 3);

            int index = 2;
            AD = BitConverter.ToUInt16(buf, index);
            

            index += 2;
            if (hasENC1)
            {
                ENC1 = BitConverter.ToUInt16(buf, index);
                index += 2;
            }

            if (hasENC2)
            {
                ENC2 = BitConverter.ToUInt16(buf, index);
                index += 2;
            }

            if (hasIn)
            {
                In = BitConverter.ToUInt16(buf, index);
                index += 2;
            }

            if (hasOut)
            {
                Out = BitConverter.ToUInt16(buf, index);
                index += 2;
            }
            if (hasAd2)
            {
                AD2 = BitConverter.ToUInt16(buf, index);
                index += 2;
            }

            ADs.Add(AD);
            AD2s.Add(AD2);
            ENC1s.Add(ENC1);
            ENC2s.Add(ENC2);
            Ins.Add(In);

            if (ADs.Count >= 200) {
                TimeGridEvent?.Invoke(this, new TimeGridEventArgs()
                {
                    ADs = ADs.ToArray(),
                    AD2s = AD2s.ToArray(),
                    ENC1s = ENC1s.ToArray(),
                    ENC2s = ENC2s.ToArray(),
                    Ins = Ins.ToArray()
                });
                ADs.Clear();
                AD2s.Clear();
                ENC1s.Clear();
                ENC2s.Clear();
                Ins.Clear();
            }
        }

        void Clear() {
            Now = DateTime.MinValue;
            ADs.Clear();
            AD2s.Clear();
            ENC1s.Clear();
            ENC2s.Clear();
            Ins.Clear();
        }
        void ParseFuncPack(List<byte> datas)
        {
            FuncRecPackEvent?.Invoke(this, new PackEventArgs() { Time = DateTime.Now, Pack = datas.ToArray() });
        }

        /// <summary>
        /// 数据发送，不需要包含开头的7E
        /// </summary>
        /// <param name="datas"></param>
        public void Send(List<byte> datas) 
        {
            datas.Insert(0, 0x80);//插入B0
            
            //需要在后面添加CRC8
            byte crc8 = Misc.CRC.CRC8(datas, 0, datas.Count);
            datas.Add(crc8);

            FuncSendPackEvent?.Invoke(this, new PackEventArgs() { Time = DateTime.Now, Pack = datas.ToArray() });
            ProtocolCommon.Data2Pdu(datas, out List<byte> pack);
            pack.Add(0x7E);//一定要加0x7E

            if (comm == null)
                return;

            comm.SendMsg(pack.ToArray());
        }
    }

    public class FlyAd2021JsonDb 
    {
        public string Addr { get; set; } = "COM1";
        public static bool Save(FlyAd2021ServiceClient flyAd2021, string filePath) {
            FlyAd2021JsonDb jsonDb = new FlyAd2021JsonDb() {
                Addr = flyAd2021.Addr,
            };
            string json = Newtonsoft.Json.JsonConvert.SerializeObject(jsonDb, Newtonsoft.Json.Formatting.Indented);
            File.WriteAllText(filePath, json);
            return true;
        }
        public static bool Load(FlyAd2021ServiceClient flyAd2021, string filePath)
        {
            if (!File.Exists(filePath)) {
                return false;
            }
            string json = File.ReadAllText(filePath);
            Newtonsoft.Json.JsonConvert.PopulateObject(json, flyAd2021);
            return true;
        }
    }
}
